diff -Naur -X diff-excludes.list dovecot-1.0.0/src/lib/network.c dovecot-1.0.0_multi_listen/src/lib/network.c
--- dovecot-1.0.0/src/lib/network.c	2006-10-16 08:00:21.000000000 +0900
+++ dovecot-1.0.0_multi_listen/src/lib/network.c	2007-04-13 21:09:05.087859675 +0900
@@ -445,9 +445,10 @@
 	if (host_error != 0)
 		return host_error;
 
-	if (getnameinfo(ai->ai_addr, ai->ai_addrlen, hbuf,
-			sizeof(hbuf), NULL, 0, NI_NUMERICHOST) != 0)
-		return 1;
+        host_error = getnameinfo(ai->ai_addr, ai->ai_addrlen, hbuf,
+                                 sizeof(hbuf), NULL, 0, NI_NUMERICHOST);
+	if (host_error != 0)
+		return host_error;
 
 
         /* get number of IPs */
diff -Naur -X diff-excludes.list dovecot-1.0.0/src/master/login-process.c dovecot-1.0.0_multi_listen/src/master/login-process.c
--- dovecot-1.0.0/src/master/login-process.c	2007-03-27 04:09:57.000000000 +0900
+++ dovecot-1.0.0_multi_listen/src/master/login-process.c	2007-04-13 21:09:05.087859675 +0900
@@ -60,7 +60,7 @@
 static bool login_process_init_group(struct login_process *p);
 static void login_processes_start_missing(void *context);
 
-static void login_group_create(struct settings *set)
+static struct login_group *login_group_create(struct settings *set)
 {
 	struct login_group *group;
 
@@ -72,6 +72,7 @@
 
 	group->next = login_groups;
 	login_groups = group;
+        return group;
 }
 
 static void login_group_unref(struct login_group *group)
@@ -213,15 +214,32 @@
 	p->state = state;
 }
 
+static void login_groups_create(struct settings *set)
+{
+        unsigned int    li, sli, n_listens, n_ssl_listens;
+	struct login_group *group;
+
+        n_listens = array_count(&set->listen_fds);
+        n_ssl_listens = array_count(&set->ssl_listen_fds);
+        for(li=0, sli=0; (li < n_listens) || (sli < n_ssl_listens);) {
+                group = login_group_create(set);
+                if(li < n_listens)
+                        group->listen_fd = *(int*)(array_idx(&set->listen_fds, li++));
+                if(sli < n_ssl_listens)
+                        group->ssl_listen_fd = *(int*)(array_idx(&set->ssl_listen_fds, sli++));
+        }
+}
+
+
 static void login_process_groups_create(void)
 {
 	struct server_settings *server;
 
 	for (server = settings_root; server != NULL; server = server->next) {
 		if (server->imap != NULL)
-			login_group_create(server->imap);
+			login_groups_create(server->imap);
 		if (server->pop3 != NULL)
-			login_group_create(server->pop3);
+			login_groups_create(server->pop3);
 	}
 }
 
@@ -634,12 +652,12 @@
 	log_set_prefix(log, prefix);
 
 	/* move the listen handle */
-	if (dup2(group->set->listen_fd, LOGIN_LISTEN_FD) < 0)
+	if (dup2(group->listen_fd, LOGIN_LISTEN_FD) < 0)
 		i_fatal("dup2(listen_fd) failed: %m");
 	fd_close_on_exec(LOGIN_LISTEN_FD, FALSE);
 
 	/* move the SSL listen handle */
-	if (dup2(group->set->ssl_listen_fd, LOGIN_SSL_LISTEN_FD) < 0)
+	if (dup2(group->ssl_listen_fd, LOGIN_SSL_LISTEN_FD) < 0)
 		i_fatal("dup2(ssl_listen_fd) failed: %m");
 	fd_close_on_exec(LOGIN_SSL_LISTEN_FD, FALSE);
 
@@ -734,6 +752,7 @@
 
 static int login_group_start_missings(struct login_group *group)
 {
+
 	if (group->set->login_process_per_connection &&
 	    group->processes >= group->set->login_max_processes_count &&
 	    group->listening_processes == 0) {
diff -Naur -X diff-excludes.list dovecot-1.0.0/src/master/login-process.h dovecot-1.0.0_multi_listen/src/master/login-process.h
--- dovecot-1.0.0/src/master/login-process.h	2006-10-14 08:20:34.000000000 +0900
+++ dovecot-1.0.0_multi_listen/src/master/login-process.h	2007-04-13 21:09:05.095859488 +0900
@@ -12,6 +12,9 @@
 	unsigned int listening_processes;
 	unsigned int wanted_processes_count;
 
+        int listen_fd;
+        int ssl_listen_fd;
+
 	/* if login_process_per_connection=yes this contains the list of
 	   processes that are in LOGIN_STATE_FULL_PRELOGINS state */
 	struct login_process *oldest_prelogin_process;
diff -Naur -X diff-excludes.list dovecot-1.0.0/src/master/main.c dovecot-1.0.0_multi_listen/src/master/main.c
--- dovecot-1.0.0/src/master/main.c	2007-03-28 05:11:42.000000000 +0900
+++ dovecot-1.0.0_multi_listen/src/master/main.c	2007-04-13 21:09:05.099859394 +0900
@@ -254,15 +254,15 @@
 }
 
 static void resolve_ip(const char *set_name, const char *name,
-		       struct ip_addr *ip, unsigned int *port)
+		       struct ip_addr **ips, unsigned int *port, int* ips_countp)
 {
 	struct ip_addr *ip_list;
 	const char *p;
 	int ret, ips_count;
 
 	if (*name == '\0') {
-                /* defaults to "*" or "[::]" */
-		ip->family = 0;
+                *ips = NULL;
+                *ips_countp = 0;
 		return;
 	}
 
@@ -298,13 +298,19 @@
 
 	if (strcmp(name, "*") == 0) {
 		/* IPv4 any */
+                struct ip_addr  *ip = t_malloc(sizeof(struct ip_addr));
 		net_get_ip_any4(ip);
+                *ips = ip;
+                *ips_countp = 1;
 		return;
 	}
 
 	if (strcmp(name, "::") == 0) {
 		/* IPv6 any */
+                struct ip_addr  *ip = t_malloc(sizeof(struct ip_addr));
 		net_get_ip_any6(ip);
+                *ips = ip;
+                *ips_countp = 1;
 		return;
 	}
 
@@ -318,23 +324,33 @@
 	if (ips_count < 1)
 		i_fatal("%s: No IPs for address: %s", set_name, name);
 
-	*ip = ip_list[0];
+	*ips = ip_list;
+        *ips_countp = ips_count;
 }
 
 static void
 check_conflicts_set(const struct settings *set, const struct ip_addr *ip,
 		    unsigned int port, const char *name1, const char *name2)
 {
-	if (set->listen_port == port && net_ip_compare(ip, &set->listen_ip) &&
-	    set->listen_fd > 0) {
-		i_fatal("Protocols %s and %s are listening in same ip/port",
-			name1, name2);
-	}
-	if (set->ssl_listen_port == port &&
-	    net_ip_compare(ip, &set->ssl_listen_ip) && set->ssl_listen_fd > 0) {
-		i_fatal("Protocols %ss and %s are listening in same ip/port",
-			name1, name2);
-	}
+        int     i, n;
+
+        for (n = array_count(&set->listen_ips), i = 0; i < n; i++) {
+                if (*(unsigned int*)(array_idx(&set->listen_ports, i)) == port &&
+                    net_ip_compare(ip, array_idx(&set->listen_ips, i)) &&
+                    *(int*)(array_idx(&set->listen_fds, i)) > 0) {
+                        i_fatal("Protocols %s and %s are listening in same ip/port",
+                                name1, name2);
+                }
+        }
+
+        for (n = array_count(&set->ssl_listen_ips), i = 0; i < n; i++) {
+                if (*(unsigned int*)(array_idx(&set->ssl_listen_ports, i)) == port &&
+                    net_ip_compare(ip, array_idx(&set->ssl_listen_ips, i)) &&
+                    *(int*)(array_idx(&set->ssl_listen_fds, i)) > 0) {
+                        i_fatal("Protocols %ss and %s are listening in same ip/port",
+                                name1, name2);
+                }
+        }
 }
 
 static void check_conflicts(const struct ip_addr *ip, unsigned int port,
@@ -356,108 +372,196 @@
 
 static void listen_protocols(struct settings *set, bool retry)
 {
-	struct ip_addr *ip;
+        array_t  *ips;
+        array_t  *ports;
+        array_t  *fds;
 	const char *const *proto;
-	unsigned int port;
-	int *fd, i;
+	const char *const *listens;
+	int fd_init = -1;
+        unsigned int    default_port, default_ssl_port, i, n_listens=0, n_ssl_listens=0, n=0 ;
 
-	set->listen_port = set->protocol == MAIL_PROTOCOL_IMAP ? 143 : 110;
+	default_port = set->protocol == MAIL_PROTOCOL_IMAP ? 143 : 110;
 #ifdef HAVE_SSL
-	set->ssl_listen_port = set->protocol == MAIL_PROTOCOL_IMAP ? 993 : 995;
+	default_ssl_port = set->protocol == MAIL_PROTOCOL_IMAP ? 993 : 995;
 #else
-	set->ssl_listen_port = 0;
+        default_ssl_port = 0;
 #endif
 
-	/* resolve */
-	resolve_ip("listen", set->listen, &set->listen_ip, &set->listen_port);
+        listens = t_strsplit_spaces(set->listen, ", ");
+        for (i = 0; listens[i] != NULL; i++);
+        n = i;
+        if (n > 0) {
+                ARRAY_CREATE(&set->listen_ips, system_pool, struct ip_addr, 0);
+                ARRAY_CREATE(&set->listen_ports, system_pool, int, 0);
+                /* resolve */
+                for (i = 0; i < n; i++) {
+                        unsigned int    tmp_port = 0;
+                        struct ip_addr  *tmp_ips = NULL;
+                        int             ips_count = 0, ip_i;
+                        resolve_ip("listen", listens[i], &tmp_ips, &tmp_port, &ips_count);
+                        if (ips_count == 0)
+                                continue;
+                        if (tmp_port == 0)
+                                tmp_port = default_port;
+                        array_append(&set->listen_ips, tmp_ips, ips_count);
+                        for (ip_i = 0; ip_i < ips_count; ip_i++) {
+                                array_append(&set->listen_ports, &tmp_port, 1);
+                        }
+                        n_listens += ips_count;
+                }
+        }
+        if (n_listens == 0) {
+                struct ip_addr  tmp_ip = { .family = 0 };
+                ARRAY_CREATE(&set->listen_ips, system_pool, struct ip_addr, 1);
+                array_idx_set(&set->listen_ips, 0, &tmp_ip);
+                ARRAY_CREATE(&set->listen_ports, system_pool, int, 1);
+                array_idx_set(&set->listen_ports, 0, &default_port);
+                n_listens = 1;
+        }
+        ARRAY_CREATE(&set->listen_fds, system_pool, int, n_listens);
+        for (i = 0; i < n_listens; i++)
+                array_idx_set(&set->listen_fds, i, &fd_init);
 	if (!set->ssl_disable) {
-		resolve_ip("ssl_listen", set->ssl_listen, &set->ssl_listen_ip,
-			   &set->ssl_listen_port);
-	}
+                const char *const *ssl_listens = NULL;
+                int                     allocated_array = 0;
+                if (*set->ssl_listen != '\0') {
+                        ssl_listens = t_strsplit_spaces(set->ssl_listen, ", ");
+                        for (i = 0; ssl_listens[i] != NULL; i++);
+                        n = i;
+                        ARRAY_CREATE(&set->ssl_listen_ips, system_pool, struct ip_addr, 0);
+                        ARRAY_CREATE(&set->ssl_listen_ports, system_pool, int, 0);
+                        allocated_array = 1;
+                        for (i = 0; i < n; i++) {
+                                unsigned int    tmp_port = 0;
+                                struct ip_addr  *tmp_ips = NULL;
+                                int             ips_count = 0, ip_i;
+                                resolve_ip("listen", ssl_listens[i], &tmp_ips, &tmp_port, &ips_count);
+                                if (ips_count == 0)
+                                        continue;
+                                if (tmp_port == 0)
+                                        tmp_port = default_ssl_port;
+                                array_append(&set->ssl_listen_ips, tmp_ips, ips_count);
+                                for (ip_i = 0; ip_i < ips_count; ip_i++) {
+                                        array_append(&set->ssl_listen_ports, &tmp_port, 1);
+                                }
+                                n_ssl_listens += ips_count;
+                        }
+                }
+                if (n_ssl_listens == 0) {
+                        if (!allocated_array) {
+                                ARRAY_CREATE(&set->ssl_listen_ips, system_pool, struct ip_addr, 0);
+                                ARRAY_CREATE(&set->ssl_listen_ports, system_pool, int, 0);
+                        }
+                        for (i = 0; i < n_listens; i++) {
+                                array_append(&set->ssl_listen_ips,
+                                             array_idx(&set->listen_ips, i), 1);
+                                array_append(&set->ssl_listen_ports, &default_ssl_port, 1);
+                                n_ssl_listens ++;
+                        }
+                }
+                ARRAY_CREATE(&set->ssl_listen_fds, system_pool, int, n_ssl_listens);
+                for (i = 0; i < n_ssl_listens; i++)
+                        array_idx_set(&set->ssl_listen_fds, i, &fd_init);
 
-	/* if ssl_listen wasn't explicitly set in the config file,
-	   use the non-ssl IP settings for the ssl listener, too. */
-	if (set->ssl_listen_ip.family == 0 && *set->ssl_listen == '\0')
-		set->ssl_listen_ip = set->listen_ip;
+        }
 
 	/* register wanted protocols */
         proto = t_strsplit_spaces(set->protocols, " ");
 	for (; *proto != NULL; proto++) {
-		fd = NULL; ip = NULL; port = 0;
+		fds = NULL; ips = NULL; ports = NULL;
 		if (strcasecmp(*proto, "imap") == 0) {
 			if (set->protocol == MAIL_PROTOCOL_IMAP) {
-				fd = &set->listen_fd;
-				port = set->listen_port;
-				ip = &set->listen_ip;
+				fds = &set->listen_fds;
+				ports = &set->listen_ports;
+				ips = &set->listen_ips;
+                                n = n_listens;
 			}
 		} else if (strcasecmp(*proto, "imaps") == 0) {
 			if (set->protocol == MAIL_PROTOCOL_IMAP &&
 			    !set->ssl_disable) {
-				fd = &set->ssl_listen_fd;
-				port = set->ssl_listen_port;
-				ip = &set->ssl_listen_ip;
+				fds = &set->ssl_listen_fds;
+				ports = &set->ssl_listen_ports;
+				ips = &set->ssl_listen_ips;
+                                n = n_ssl_listens;
 			}
 		} else if (strcasecmp(*proto, "pop3") == 0) {
 			if (set->protocol == MAIL_PROTOCOL_POP3) {
-				fd = &set->listen_fd;
-				port = set->listen_port;
-				ip = &set->listen_ip;
+				fds = &set->listen_fds;
+				ports = &set->listen_ports;
+				ips = &set->listen_ips;
+                                n = n_listens;
 			}
 		} else if (strcasecmp(*proto, "pop3s") == 0) {
 			if (set->protocol == MAIL_PROTOCOL_POP3 &&
 			    !set->ssl_disable) {
-				fd = &set->ssl_listen_fd;
-				port = set->ssl_listen_port;
-				ip = &set->ssl_listen_ip;
+				fds = &set->ssl_listen_fds;
+				ports = &set->ssl_listen_ports;
+				ips = &set->ssl_listen_ips;
+                                n = n_ssl_listens;
 			}
 		} else {
 			i_fatal("Unknown protocol %s", *proto);
 		}
 
-		if (fd == NULL)
+		if (fds == NULL)
 			continue;
 
-		if (*fd != -1)
-			i_fatal("Protocol %s given more than once", *proto);
-
-		if (port == 0)
-			*fd = null_fd;
-		else {
-			for (i = 0; i < 10; i++) {
-				*fd = net_listen(ip, &port, 8);
-				if (*fd != -1)
-					break;
-				if (errno == EADDRINUSE) {
-					/* retry */
-				} else if (errno == EINTR &&
-					   io_loop_is_running(ioloop)) {
-					/* SIGHUPing sometimes gets us here.
-					   we don't want to die. */
-				} else {
-					/* error */
-					break;
-				}
-
-				check_conflicts(ip, port, *proto);
-				if (!retry)
-					break;
-
-				/* wait a while and try again. we're SIGHUPing
-				   so we most likely just closed it ourself.. */
-				sleep(1);
-			}
-
-			if (*fd == -1)
-				i_fatal("listen(%d) failed: %m", port);
-			net_set_nonblock(*fd, TRUE);
-			fd_close_on_exec(*fd, TRUE);
-		}
-	}
+                for (i = 0 ; i < n; i++) {
+                        int     *fd = array_idx_modifyable(fds, i);
+                        unsigned int    *port = array_idx_modifyable(ports, i);
+
+                        if (*fd != -1)
+                                i_fatal("Protocol %s given more than once", *proto);
+
+                        if (*port == 0)
+                                *fd = null_fd;
+                        else {
+                                struct ip_addr  *ip = array_idx_modifyable(ips, i);
+                                int     j;
+                                for (j = 0; j < 10; j++) {
+                                        *fd = net_listen(ip, port, 8);
+                                        if (*fd != -1)
+                                                break;
+                                        if (errno == EADDRINUSE) {
+                                                /* retry */
+                                        } else if (errno == EINTR &&
+                                                   io_loop_is_running(ioloop)) {
+                                                /* SIGHUPing sometimes gets us here.
+                                                   we don't want to die. */
+                                        } else {
+                                                /* error */
+                                                break;
+                                        }
+
+                                        if (i > 0)
+                                                check_conflicts(ip, *port, *proto);
+                                        if (!retry)
+                                                break;
+
+                                        /* wait a while and try again. we're SIGHUPing
+                                           so we most likely just closed it ourself.. */
+                                        sleep(1);
+                                }
+
+                                if (*fd == -1) {
+                                        i_fatal("listen(%d) failed: %m", *port);
+                                        *fd = null_fd;
+                                }else{
+                                        net_set_nonblock(*fd, TRUE);
+                                        fd_close_on_exec(*fd, TRUE);
+                                }
+                        }
+                }
+	}
+
+        for (i = 0; i < n_listens; i++)
+                if(*(int*)(array_idx(&set->listen_fds, i)) == -1)
+                        array_idx_set(&set->listen_fds, i, &null_fd);
+
+        for (i = 0; i < n_ssl_listens; i++)
+                if(*(int*)(array_idx(&set->ssl_listen_fds, i)) == -1)
+                        array_idx_set(&set->ssl_listen_fds, i, &null_fd);
 
-	if (set->listen_fd == -1)
-		set->listen_fd = null_fd;
-	if (set->ssl_listen_fd == -1)
-		set->ssl_listen_fd = null_fd;
 }
 
 static void listen_fds_open(bool retry)
@@ -472,24 +576,36 @@
 	}
 }
 
+static void listen_fds_array_close(struct settings *set)
+{
+        int      fd, invalid_fd=-1;
+        unsigned int i;
+        unsigned int n_listens = array_count(&set->listen_fds);
+        unsigned int n_ssl_listens = array_count(&set->ssl_listen_fds);
+
+        for (i = 0; i < n_listens; i++) {
+                if((fd = *(int*)(array_idx(&set->listen_fds, i))) == null_fd)
+                        if(close(fd) < 0)
+				i_error("close(imap.listen_fd) failed: %m");
+                array_idx_set(&set->listen_fds, i, &invalid_fd);
+        }
+
+        for (i = 0; i < n_ssl_listens; i++) {
+                if((fd = *(int*)(array_idx(&set->listen_fds, i))) == null_fd)
+                        if(close(fd) < 0)
+				i_error("close(imap.ssl_listen_fd) failed: %m");
+                array_idx_set(&set->listen_fds, i, &invalid_fd);
+        }
+}
+
 static void listen_fds_close(struct server_settings *server)
 {
 	for (; server != NULL; server = server->next) {
 		if (server->imap != NULL) {
-			if (server->imap->listen_fd != null_fd &&
-			    close(server->imap->listen_fd) < 0)
-				i_error("close(imap.listen_fd) failed: %m");
-			if (server->imap->ssl_listen_fd != null_fd &&
-			    close(server->imap->ssl_listen_fd) < 0)
-				i_error("close(imap.ssl_listen_fd) failed: %m");
+                        listen_fds_array_close(server->imap);
 		}
 		if (server->pop3 != NULL) {
-			if (server->pop3->listen_fd != null_fd &&
-			    close(server->pop3->listen_fd) < 0)
-				i_error("close(pop3.listen_fd) failed: %m");
-			if (server->pop3->ssl_listen_fd != null_fd &&
-			    close(server->pop3->ssl_listen_fd) < 0)
-				i_error("close(pop3.ssl_listen_fd) failed: %m");
+                        listen_fds_array_close(server->pop3);
 		}
 	}
 }
diff -Naur -X diff-excludes.list dovecot-1.0.0/src/master/master-settings.c dovecot-1.0.0_multi_listen/src/master/master-settings.c
--- dovecot-1.0.0/src/master/master-settings.c	2007-03-28 06:35:33.000000000 +0900
+++ dovecot-1.0.0_multi_listen/src/master/master-settings.c	2007-04-13 21:09:05.099859394 +0900
@@ -266,8 +266,6 @@
 	MEMBER(pop3_logout_format) "top=%t/%p, retr=%r/%b, del=%d/%m, size=%s",
 
 	/* .. */
-	MEMBER(listen_fd) -1,
-	MEMBER(ssl_listen_fd) -1
 };
 
 struct auth_settings default_auth_settings = {
diff -Naur -X diff-excludes.list dovecot-1.0.0/src/master/master-settings.h dovecot-1.0.0_multi_listen/src/master/master-settings.h
--- dovecot-1.0.0/src/master/master-settings.h	2007-03-28 06:35:11.000000000 +0900
+++ dovecot-1.0.0_multi_listen/src/master/master-settings.h	2007-04-13 21:09:05.099859394 +0900
@@ -2,6 +2,7 @@
 #define __MASTER_SETTINGS_H
 
 #include "network.h"
+#include "array.h"
 
 enum mail_protocol {
         MAIL_PROTOCOL_ANY,
@@ -122,12 +123,22 @@
 	const char *pop3_logout_format;
 
 	/* .. */
-	int listen_fd, ssl_listen_fd;
-
 	uid_t login_uid;
 
+
+        array_t ARRAY_DEFINE(listen_fds,        int);
+        array_t ARRAY_DEFINE(ssl_listen_fds,    int);
+
+        array_t ARRAY_DEFINE(listen_ips,        struct ip_addr);
+        array_t ARRAY_DEFINE(ssl_listen_ips,    struct ip_addr);
+        array_t ARRAY_DEFINE(listen_ports,      unsigned int);
+        array_t ARRAY_DEFINE(ssl_listen_ports,  unsigned int);
+
+        /*
 	struct ip_addr listen_ip, ssl_listen_ip;
 	unsigned int listen_port, ssl_listen_port;
+        */
+
 
 	const char *imap_generated_capability;
 
